/**
 * @file ldata.c
 *
 * @brief Wrapper to handle reading/writing the ldata file.
 *
 * Optimizes to minimize the opens and frees, plus tries to read from the
 * filesystem instead of always looking for a packfile.
 */

#include "lephisto.h"
#include "log.h"
#include "md5.h"
#include "lxml.h"
#include "pack.h"
#include "lfile.h"
#include "ldata.h"

#define LDATA_FILENAME  "ldata"
#define LDATA_DEF       "ldata"

#define XML_START_ID    "Start" /**< XML document tag of module start file. */
#define START_DATA      "../dat/start.xml" /**< Path to module start file. */

/* Packfile. */
static char* ldata_filename   = NULL; /**< Packfile name. */
static Packfile_t* ldata_pack = NULL; /**< Actual packfile. */
static char* ldata_packName   = NULL; /**< Name of the ldata module. */

/* File list. */
static char** ldata_fileList    = NULL; /**< List of the files in the packfile. */
static uint32_t ldata_fileNList = 0;    /**< Number of files in ldata_fileList. */

/**
 * @brief Check to see if path is a ldata file.
 *    @param path Path to check to see if it's an ldata file.
 *    @return 1 if it is an ldata file, otherwise 0.
 */
int ldata_check(char* path) {
  return pack_check(path);
}

/**
 * @brief Set the current ldata path to use.
 *    @param path Path to set.
 */
int ldata_setPath(char* path) {
  if(ldata_filename != NULL)
    free(ldata_filename);
  ldata_filename = (path == NULL) ? NULL : strdup(path);
  return 0;
}

/**
 * @brief Open a packfile if needed.
 */
static int ldata_openPackfile(void) {
  int i;
  char** files;
  int nfiles;
  size_t len;

  if(lfile_fileExists("%s-%d.%d.%d", LDATA_FILENAME, VMAJOR, VMINOR, VREV)) {
    ldata_filename = malloc(PATH_MAX);
    snprintf(ldata_filename, PATH_MAX, "%s-%d.%d.%d", LDATA_FILENAME, VMAJOR,
        VMINOR, VREV);
  }
  else if(lfile_fileExists(LDATA_DEF))
    ldata_filename = strdup(LDATA_DEF);
  else {
    files = lfile_readDir(&nfiles, ".");
    len = strlen(LDATA_FILENAME);
    for(i = 0; i < nfiles; i++) {
      if(strncmp(files[i], LDATA_FILENAME, len)==0) {
        /* Must be packed. */
        if(pack_check(files[i]))
          continue;

        ldata_filename = strdup(files[i]);
        break;
      }
    }
    for(i = 0; i < nfiles; i++)
      free(files[i]);
    free(files);
  }
  return 0;
}

/**
 * @brief Open the ldata file.
 *    @return 0 on success.
 */
int ldata_open(void) {
  return 0;
}

/**
 * @brief Close and clean up the ldata file.
 */
void ldata_close(void) {
  unsigned int i;

  /* Destroy the name. */
  if(ldata_packName != NULL) {
    free(ldata_packName);
    ldata_packName = NULL;
  }

  /* Destroy the list. */
  if(ldata_fileList != NULL) {
    for(i = 0; i < ldata_fileNList; i++)
      free(ldata_fileList[i]);

    free(ldata_fileList);
    ldata_fileList = NULL;
    ldata_fileNList = 0;
  }

  /* Close the packfile. */
  /*
  pack_close(ldata_pack);
  ldata_pack = NULL;
  */
}

/**
 * @brief Get the ldata's name.
 *    @return The ldata's name.
 */
char* ldata_name(void) {
  char* buf;
  uint32_t size;
  xmlNodePtr node;
  xmlDocPtr doc;

  /* Alreay loaded. */
  if(ldata_packName != NULL)
    return ldata_packName;

  /* We'll just read it and parse it. */
  buf = ldata_read(START_DATA, &size);
  doc = xmlParseMemory(buf, size);

  /* Make sure it's what we are looking for. */
  node = doc->xmlChildrenNode;
  if(!xml_isNode(node, XML_START_ID)) {
    ERR("Malformed '"START_DATA"' file: missing root element '"XML_START_ID"'");
    return NULL;
  }

  /* Check if node is valid. */
  node = node->xmlChildrenNode; /* First system node. */
  if(node == NULL) {
    ERR("Malformed '"START_DATA"' file: does not contain elements");
    return NULL;
  }
  do {
    xmlr_strd(node, "name", ldata_packName);
  } while(xml_nextNode(node));

  xmlFreeDoc(doc);
  free(buf);

  /* Check if data name is found. */
  if(ldata_packName == NULL)
    WARN("No ldata packname found.");
  
  return ldata_packName;
}

/**
 * @brief Read a file from the ldata.
 */
void* ldata_read(const char* filename, uint32_t* filesize) {
  if(ldata_filename == NULL)
    ldata_openPackfile();
  return pack_readfile(ldata_filename, filename, filesize);
}

/**
 * @brief Get the list of files in the ldata.
 */
char** ldata_list(const char* path, uint32_t* nfiles) {
  (void)path;

  if(ldata_fileList != NULL) {
    *nfiles = ldata_fileNList;
    return ldata_fileList;
  }

  if(ldata_filename == NULL)
    ldata_openPackfile();

  ldata_fileList = pack_listfiles(ldata_filename, &ldata_fileNList);
  *nfiles = ldata_fileNList;
  return ldata_fileList;
}

